import machine, uasyncio, config

LOOPTIME = 50

# VEML7700 ambient brightness sensor

# rp2 PICO_W
# id=0, scl=Pin(5), sda=Pin(4)

# esp32 GENERIC_S3
# id=0, scl=Pin(9), sda=Pin(8)
# id=1, scl=Pin(25), sda=Pin(26) ???

averageBrightness = 1.0
rawBrightness = 0.0
ambientBrightness = 1.0
async def main():
	global rawBrightness, averageBrightness, ambientBrightness

	# print('BrightnessSensor')
	address = 0x10
	if config.Module == 'esp8266': i2c = machine.I2C(sda=machine.Pin(4), scl=machine.Pin(5), freq=400000)
	elif config.Module == 'esp32': i2c = machine.I2C(0, freq=400000, sda=machine.Pin(7), scl=machine.Pin(0))
	else: pass

	if config.AmbientLightSensor == 'veml7700':
		data = (0b01 << 11) + (0b0011<<6) + (0b00 << 4) + (0b0 << 1) + (0b0 << 0)	# Gain x1, IT 800ms, PERS 1, INT disable, Power On
		i2c.writeto(address, bytes([0, data & 0xff, data >> 8]))	# Command Code: ALS_CONF_0, LSB, MSB
	elif config.AmbientLightSensor == 'veml3328':
		# SD1 on, SD_ALS GCIR, DG x1, GAIN x4, SENS high, IT 400ms, AF auto, TRIG no, SD0 On
		data = (0b0 << 15) + (0b1 << 14) + (0b00 << 12) + (0b11 << 10) + (0b0 << 6) + (0b11<<4) + (0b0 << 3) + (0b0 << 2) + (0b0 << 0)	
		i2c.writeto(address, bytes([0, data & 0xff, data >> 8]))	# Command Code: ALS_CONF_0, LSB, MSB
	else: pass

	while True:
		i2c.writeto(address, bytes([4]), False)	# Command Code: ALS
		data = i2c.readfrom(address, 2)
		rawBrightness = (data[0] + (data[1] << 8)) / 65535.0
		averageBrightness = (averageBrightness * 63.0 + rawBrightness) / 64.0
		ambientBrightness = min(max(0, pow(averageBrightness, config.Gamma) * config.Brightness), 1.0)
		# print('BrightnessSensor.sleep')
		await uasyncio.sleep_ms(LOOPTIME)

if __name__ == '__main__': uasyncio.run(main())
