import network, uasyncio, esp32, machine, os, errno, socket, json, gc, hashlib, binascii, struct, shutil, utarfile, config, language, static
from wordclock import dotmatrix

networks, scan_start, values, valid = [], uasyncio.Event(), {}, False

# based on https://github.com/p-doyle/Micropython-DNSServer-Captive-Portal
async def name_server(buffersize=128):
	print('CaptivePortal.name_server')
	server_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
	server_socket.setblocking(False)
	server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
	server_socket.bind(('0.0.0.0', 53))
	data = memoryview(bytearray(buffersize))
	while True:
		gc.collect()
		try:
			#yield uasyncio.core._io_queue.queue_read(server_socket)
			datalen, addr = server_socket.recvfrom_into(data)
			tipo, ini, lon, domain = (data[2] >> 3) & 15, 12, data[12], ''
			while tipo == 0 and lon != 0:
				domain += str(data[ini + 1:ini + lon + 1], 'utf-8') + '.'
				ini += lon + 1
				lon = data[ini]
			if not domain: continue
			print(f'CaptivePortal.name_server {addr} {datalen} {domain}')
			data[2:4] = b'\x81\x80'
			data[6:8] = data[4:6]
			data[8:12] = b'\x00\x00\x00\x00'   # Questions and Answers Counts
			data[datalen:datalen+2] = b'\xC0\x0C'  # Pointer to domain name
			data[datalen+2:datalen+12] = b'\x00\x01\x00\x01\x00\x00\x00\x3C\x00\x04'  # Response type, ttl and resource data length -> 4 bytes
			data[datalen+12:datalen+16] = bytes(map(int, config.CaptivePortalIP.split('.')))  # 4bytes of IP
			server_socket.sendto(data[:datalen+16], addr)
		except Exception as exception:
			if exception.errno == errno.EAGAIN: await uasyncio.sleep_ms(100)
			else: print(f'CaptivePortal.name_server {exception}')
	server_socket.close()
	print('CaptivePortal.name_server done')

async def sendfile(writer, filename, buffersize=1024):
	print(f'CaptivePortal.sendfile {filename}')
	writer.write(b'HTTP/1.0 200 OK\r\n\r\n')
	await writer.drain()
	attrname = filename.replace('/', '_').replace('.', '_')
	try:
		if hasattr(static, attrname):
			writer.out_buf = getattr(static, attrname)
			await writer.drain()
		else:
			buffer = memoryview(bytearray(buffersize))
			filesize = 0
			with open(filename, 'rb') as file:
				while(bytes_read := file.readinto(buffer)):
					#print(f'CaptivePortal.sendfile {filename} {bytes_read}')
					writer.write(buffer[:bytes_read])
					filesize += bytes_read
					await writer.drain()
			del(buffer)
			values['Status'] = f'sendfile {filename} {filesize}'
			print(values['Status'])
	except Exception as exception: values['Status'] = f'sendfile {filename} {exception}'

async def index(reader, writer): await sendfile(writer, 'static/index.html')

async def favicon(reader, writer): await sendfile(writer, 'static/favicon.ico')

async def font(reader, writer): await sendfile(writer, 'font.woff')

async def ota(reader, writer):
	print(f'CaptivePortal.ota {reader.method=} {reader.path=} {reader.query=} {reader.headers=}')

	writer.write('HTTP/1.1 200 OK\r\nContent-Type: text/plain\r\nTransfer-Encoding: chunked\r\n\r\n'.encode())
	await writer.drain()

	content_length = int(reader.headers['content-length'])
	filename = reader.headers['filename']
	if filename.endswith('.tar'):
		values['Status'] = f'ota updating WordClock with {filename}'
		buffer = memoryview(bytearray(1024))
		write_size = 0
		with open(filename, 'wb') as file:
			while write_size < content_length:
				write_size += (read_size := await reader.readinto(buffer))
				if read_size == None:
					values['Status'] = f'ota premature eof! {content_length=} > {write_size}'
					return
				file.write(buffer[:read_size])
			file.flush()

		values['Status'] = f'ota extracting {filename}'
		await uasyncio.sleep(1)
		tarfile = utarfile.TarFile(filename)
		for item in tarfile:
			if item.type == utarfile.DIRTYPE:
				dirname = item.name.rstrip('/')
				values['Status'] = f'ota saving {dirname}'
				try: shutil.rmtree(f'{dirname}.bak')
				except: pass
				try: os.rename(dirname, f'{dirname}.bak')
				except: pass
				os.mkdir(dirname)
			else:
				values['Status'] = f'ota updating {item.name} {item.size}'
				with open(item.name, 'wb') as file:
					shutil.copyfileobj(tarfile.extractfile(item), file)
					file.flush()
			await uasyncio.sleep(0.5)
		os.sync()
		values['Status'] = f'ota succeeded. Rebooting now...'
		await uasyncio.sleep(3)
		machine.reset()
		return

	elif filename.endswith('.bin'):
		current_partition = esp32.Partition(esp32.Partition.RUNNING)
		next_partition = current_partition.get_next_update()
		sector_count = next_partition.ioctl(4, None)
		sector_size = next_partition.ioctl(5, None)
		partition_size = sector_count * sector_size
		values['Status'] = f'ota updating partition {next_partition} with {filename}'
		if content_length > partition_size:
			values['Status'] = f'ota failed {content_length=} > {partition_size=}'
			return

		buffer = memoryview(bytearray(sector_size))
		write_size, view_size, sector = 0, 0, 0
		while write_size < content_length:
			while view_size < sector_size and (write_size + view_size) < content_length:
				read_size = await reader.readinto(buffer[view_size:])
				if read_size == None:
					values['Status'] = f'ota premature eof! {content_length=} > {write_size + view_size=}'
					return
				view_size += read_size

			for item in range(view_size, sector_size): buffer[item] = 0xff
			next_partition.writeblocks(sector, buffer)
			write_size, view_size, sector = write_size + view_size, 0, sector + 1
			values['Status'] = f'ota {write_size} of {content_length} written'
		del(buffer)
		next_partition.ioctl(3, None)
		values['Status'] = f'Setting new boot partition...'
		await uasyncio.sleep(3)
		next_partition.set_boot()
		values['Status'] = f'Rebooting into new partition now...'
		await uasyncio.sleep(3)
		machine.reset()
	else:
		values['Status'] = f'ota unknown filetype {filename}!'

# https://www.rfc-editor.org/rfc/rfc6455#section-5.2
async def web_socket_handler(reader, writer, buffersize=256):
	global valid
	print(f'CaptivePortal.web_socket_handler {reader.get_extra_info("peername")}')
	key = reader.headers['sec-websocket-key'].encode() + b'258EAFA5-E914-47DA-95CA-C5AB0DC85B11'
	x = binascii.b2a_base64(hashlib.sha1(key).digest()).strip()
	writer.write(b'HTTP/1.1 101 Switching Protocols\r\nUpgrade: websocket\r\nConnection: Upgrade\r\nSec-WebSocket-Accept: ' + x + b'\r\n\r\n')
	await writer.drain()

	OP_TYPES = { 0x0: 'cont', 0x1: 'text', 0x2: 'bytes', 0x8: 'close', 0x9: 'ping', 0xa: 'pong' }
	buffer = memoryview(bytearray(buffersize))
	while True:
		packet_size = await reader.readinto(buffer)
		print(f'CaptivePortal.web_socket_handler {packet_size}')
		if packet_size < 2: break
		op, n, bytes_read = buffer[0], buffer[1], 2
		event_fin = bool(op & (1 << 7))
		op = op & 0x0f
		if op not in OP_TYPES: raise None
		event_type = OP_TYPES[op]
		event_masked = bool(n & (1 << 7))
		n = n & 0x7f
		if n == 126: [n], bytes_read = struct.unpack('!H', buffer[bytes_read:bytes_read+2]), bytes_read + 2
		elif n == 127: [n], bytes_read = struct.unpack('!Q', buffer[bytes_read:bytes_read+8]), bytes_read + 8
		if event_masked: mask, bytes_read = buffer[bytes_read:bytes_read+4], bytes_read + 4
		event_data, bytes_read = buffer[bytes_read:bytes_read+n], bytes_read + n
		if event_masked:
			for i in range(n): event_data[i] ^= mask[i % 4]
		if event_type == 'close': break
		elif event_type == 'text':
			event_data = str(event_data, 'utf8')
			print(f'CaptivePortal.web_socket_handler {event_type} {event_fin} {event_masked} {event_data}')
			if '=' in event_data:
				key, _separator, value = event_data.partition('=')
				key, value = key.strip(), value.strip()
				if key == 'datetime':
					year, month, day, hours, minutes, seconds = json.loads(value)
					machine.RTC().datetime((year, month, day, -1, hours, minutes, seconds, 0))	# year, month, day, weekday, hours, minutes, seconds, subseconds
					valid = True
				elif hasattr(config, key):
					try:
						if type(getattr(config, key)) == int: value = int(value)
						elif type(getattr(config, key)) == float: value = float(value)
						elif type(getattr(config, key)) == bool: value = value == 'true'
					except: pass
					values['Status'] = f'{key} = {repr(value)}'
					values[key] = value
					setattr(config, key, value)
			elif event_data == 'Save':
				open(config.__file__, 'w').write('\n'.join([ f"{key} = {repr(value)}" for key, value in config.__dict__.items() if not key.startswith('__')]))
				values['Status'] = 'config saved.'
			elif event_data == 'Scan':
				scan_start.set()
				values['Status'] = 'scan started.'
			elif event_data == 'Reboot':
				values['Status'] = 'Rebooting now...'
				await uasyncio.sleep(3)
				machine.reset()
			dotmatrix.refresh += 1
		gc.collect()

async def events_handler(reader, writer):
	async def send(msg, event):
		writer.write(f'event: {event}\r\ndata: {msg}\r\n\r\n'.encode())
		await writer.drain()

	writer.write(b'HTTP/1.0 200 OK\r\nContent-Type: text/event-stream\r\nCache-Control: no-cache\r\nConnection: keep-alive\r\nAccess-Control-Allow-Origin: *\r\n\r\n')
	await writer.drain()

	print(f'CaptivePortal.events_handler {reader.get_extra_info("peername")}')
	Values = { key : getattr(config, key) for key in ['Font', 'FaceColor', 'TextColor', 'Brightness', 'Darkness', 'Gamma', 'FadeTime', 'TimeZone', 'TimeSpeed', 'SleepStart', 'SleepEnd', 'WiFiName', 'Radio', 'Hostname']}
	Values['Status'] = f'{config.Hostname} ready. (WordClock v.{config.Version}; Micropython {os.uname().version}; {os.uname().machine})'
	try: await send(json.dumps(Values), event='Values')
	except: return
	Language = { 'Numbers' : config.LedNumbers, 'FrontText' : language.frontText }
	try: await send(json.dumps(Language), event='Language')
	except: return

	last_refresh, last_networks = -1, []
	while True:
		if last_refresh != dotmatrix.refresh:
			print(f'CaptivePortal.events_handler refresh {dotmatrix.refresh}')
			try: await send(json.dumps([bool(dot) for dot in dotmatrix.newDots]), event='Dots')
			except: break
			last_refresh = dotmatrix.refresh
		elif networks != last_networks:
			print(f'CaptivePortal.events_handler scan {networks}')
			try: await send(json.dumps(networks), event='Networks')
			except: break
			last_networks = networks
		elif values:
			print(f'CaptivePortal.events_handler values {values}')
			try: await send(json.dumps(values), event='Values')
			except: break
			values.clear()
		else: await uasyncio.sleep_ms(100)
		gc.collect()

async def scan():
	global networks
	print(f'CaptivePortal.scan')
	sta = network.WLAN(network.STA_IF)
	while True:
		if scan_start.is_set():
			if not sta.active(): sta.active(True)
			try:
				networks = [{ 'ssid' : _network[0].decode(), 'bssid' : _network[1], 'channel' : _network[2], 'rssi' : _network[3], 'security' : _network[4], 'hidden' : _network[5] } for _network in sta.scan()]
				values['Status'] = 'scan done.'
			except Exception as exception: values['Status'] = f'scan {exception}'
			scan_start.clear()
		else: await uasyncio.sleep_ms(100)
		gc.collect()
	print(f'CaptivePortal.scan done')

handlers = [
	('/', ['GET'], index),
	('/favicon.ico', ['GET'], favicon),
	('/font.woff', ['GET'], font),
	('/ota', ['PUT'], ota),
	('/ws', ['GET'], web_socket_handler),
	('/es', ['GET'], events_handler),
]

# https://developer.mozilla.org/en-US/docs/Web/HTTP/Authentication
# https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/WWW-Authenticate
async def web_server_cb(reader, writer):
	try:
		line = await reader.readline()
		if not line: raise ValueError
		parts = line.decode().split()
		if len(parts) < 3: raise ValueError
		reader.method, reader.path = parts[0], parts[1]
		parts = reader.path.split('?', 1)
		if len(parts) < 2: reader.query = None
		else: reader.path, reader.query = parts[0], parts[1]
		reader.headers = {}
		while line := (await reader.readline()).decode():
			if line == '\r\n': break
			key, _separator, value = line.partition(':')
			reader.headers[key.lower()] = value.strip()
	except: return await writer.wait_closed()

	for path, methods, handler in handlers:
		if reader.path != path or reader.method not in methods: continue
		print(f'CaptivePortal.web_server {reader.get_extra_info("peername")}, {reader.method}, {reader.path}')
		await handler(reader, writer)
		return await writer.wait_closed()

	print(f'CaptivePortal.redirect {reader.get_extra_info("peername")}, {reader.method}, {reader.path}')
	writer.write(f'HTTP/1.1 301 Moved Permanently\r\nLocation: http://{config.Hostname}.local/\r\n\r\n')
	await writer.drain()
	await writer.wait_closed()

async def ap_state():
	ap = network.WLAN(network.AP_IF)
	while True:
		if config.Radio == 'WiFi' and not ap.active():
			print('CaptivePortal ap.active')
			ap.active(True)
			ap.ifconfig((config.CaptivePortalIP, config.CaptivePortalSubnet, config.CaptivePortalIP, config.CaptivePortalIP))
			ap.config(essid=config.CaptivePortalSSID, password=config.Password, authmode=network.AUTH_WPA2_PSK)	# AUTH_WPA2_PSK AUTH_OPEN
			print(f'CaptivePortal {config.CaptivePortalSSID} {config.Password}')
		elif not config.Radio == 'WiFi' and ap.active():
			print('CaptivePortal ap.inactive')
			ap.active(False)
		await uasyncio.sleep_ms(100)

async def main():
	print('CaptivePortal')
	web_server = await uasyncio.start_server(web_server_cb, '0.0.0.0', 80)
	await uasyncio.gather(scan(), web_server.task, name_server(), ap_state())

if __name__ == '__main__': uasyncio.run(main())
