import machine, network, uasyncio, time, random, struct, sys, ussl, usocket, gc, config

TZLEN, IDLEN, NONCELEN, INTLEN, WORDLEN, VERSION, LOOPTIME = 32, 16, 4, 4, 2, 1, 5

# https://github.com/StevenRuest/async_uurequests/blob/master/async_uurequests.py

valid = False
async def main():
	global valid

# 	print('ChrigusTimeProtocol')
	sta = network.WLAN(network.STA_IF)
	rtc = machine.RTC()
	network_status = { getattr(network, status) : status for status in dir(network) if status.startswith('STAT')}
	start = time.time()
	while True:
		gc.collect()
		await uasyncio.sleep(LOOPTIME)
		if config.Radio == 'WiFi' and config.WiFiName and config.WiFiPassword:
			if valid and (time.time() - start) < (config.TimeInterval * 60): continue
			print(f'ctp {sta.active()=} {sta.isconnected()=}')
			try:
				if not sta.active():
					print('ctp not sta.active')
					sta.active(True)
					sta.config(hostname=config.Hostname)
					sta.config(reconnects=3)
				elif not sta.isconnected():
					print('ctp not sta.isconnected')
					sta.connect(config.WiFiName, config.WiFiPassword)
					while not sta.isconnected() and sta.status() == network.STAT_CONNECTING: await uasyncio.sleep_ms(100)
					if sta.status() != network.STAT_GOT_IP:
						print(f'ctp {sta.status()} != network.STAT_GOT_IP {network_status[sta.status()]}')
						sta.disconnect()
						sta.active(False)
					else: print(f'ctp STAT_GOT_IP {sta.ifconfig()[0]}')
				else:
					nonce = random.getrandbits(NONCELEN*8)
					clientid = bytes.fromhex(config.ClientID)
					data = struct.pack(f'!H{TZLEN}s{IDLEN}sI', VERSION, config.TimeZone, clientid, nonce)
					print(f'ctp > {config.TimeServer=}, {config.TimePort=}, {config.TimeZone=}, {config.TimeInterval=}, {config.Fingerprint=}, {config.ClientID=}, {VERSION=}, {nonce=}')
	
					addrinfo = usocket.getaddrinfo(config.TimeServer, config.TimePort, 0, usocket.SOCK_STREAM)[0]
					ctp_socket = usocket.socket(addrinfo[0], usocket.SOCK_STREAM, addrinfo[2])
					ctp_socket.connect(addrinfo[-1])
					gc.collect()
					ctp_socket = ussl.wrap_socket(ctp_socket, server_hostname=config.TimeServer)
					yield uasyncio.core._io_queue.queue_write(ctp_socket)
					ctp_socket.write(b"POST / HTTP/1.0\r\n")
					ctp_socket.write(b"Host: %s\r\n" % config.TimeServer)
					ctp_socket.write(b'Content-Type: application/octet-stream\r\n')
					ctp_socket.write(b"Content-Length: %d\r\n" % len(data))
					ctp_socket.write(b"Connection: close\r\n\r\n")
					yield uasyncio.core._io_queue.queue_write(ctp_socket)
					ctp_socket.write(data)
					yield uasyncio.core._io_queue.queue_read(ctp_socket)
					status_line = ctp_socket.readline()
					yield uasyncio.core._io_queue.queue_read(ctp_socket)
					while (line := ctp_socket.readline()) and line != b"\r\n": yield uasyncio.core._io_queue.queue_read(ctp_socket)
					yield uasyncio.core._io_queue.queue_read(ctp_socket)
					data = ctp_socket.read()
					ctp_socket.close()
	
					if status_line.split()[1] != b'200':
						print(f"ctp {status_line.split()}")
						continue
	
					if len(data) != IDLEN + NONCELEN + INTLEN + INTLEN + WORDLEN + WORDLEN + INTLEN: raise ValueError(f'wrong length: {len(data)}, {data=}\n')
	
					ClientID, NONCE, TimeZone, AltZone, Flags, HightTime, Time = struct.unpack(f'!{IDLEN}sIiiHHI', data)
					DST = Flags & 1 == 1
					if ClientID != clientid or NONCE != nonce:
						print(f'{ClientID} != {clientid} or {NONCE} != {nonce}')
						continue
	
					ctpTime = Time - (AltZone if DST else TimeZone);
					#print(f'ctp < {ClientID.hex()=}, {NONCE=}, {TimeZone=}, {AltZone=}, {DST=}, {HightTime=}, {Time=}, {ctpTime=}')
	
					ctpTime -= int((time.localtime(0)[0]-1970)*365.25)*24*60*60	# Epoch: Unix 1970-01-01, ESP32 2000-01-01
					tm = time.localtime(ctpTime)	# year, month, mday, hour, minute, second, weekday, yearday
					rtc.datetime((tm[0], tm[1], tm[2], tm[6] + 1, tm[3], tm[4], tm[5], 0))	# year, month, day, weekday, hours, minutes, seconds, subseconds
					valid, start = True, time.time()
					dt = time.localtime()
					print(f'ctp < {dt[0]:04d}.{dt[1]:02d}.{dt[2]:02d} {dt[3]:02d}:{dt[4]:02d}:{dt[5]:02d}')
			except Exception as exception:
				sys.print_exception(exception)
				sta.disconnect()
		elif sta.active(): sta.active(False)

if __name__ == '__main__': uasyncio.run(main())
