import time, uasyncio, gc, config, language
from wordclock import ledstrip, brightnesssensor, currenttimeservice, chrigustimeprotocol, captiveportal, dotmatrix

LOOPTIME = 50

async def main():
	print('WordClock')
	gc.collect(); last, current = 0, gc.mem_free(); print(current, 0)

	dotmatrix.add_dots(language.noTimeNumbers[0])
	dotmatrix.print_dots(0, 0)

	ledstrip.clear_strip()
	print('Ledtest start')
	for dot in range(dotmatrix.length+2):
		brightness = brightnesssensor.ambientBrightness
		ledstrip.colorFrame(dot, 0)
		ledstrip.colorFrame(dot+1, brightness)
		ledstrip.write()
		await uasyncio.sleep_ms(100)
	print('Ledtest done')

	dotmatrix.print_dots(brightness, 0)

	lastMinute5 = -1
	fadeTime = 0
	while True:
		seconds = (time.time() * config.TimeSpeed + 150) % (24*60*60)
		hour = seconds // (60*60)
		minute = seconds % (60*60) // 60
		minute5 = minute // 5
		now = f'{hour:02d}:{minute:02d}'

		if (currenttimeservice.valid or chrigustimeprotocol.valid or captiveportal.valid) and minute5 != lastMinute5:
			lastMinute5 = minute5
			fadeTime = config.FadeTime
			print(f'{brightness=} {hour=}, {minute=}, {now=}')
			gc.collect(); last, current = current, gc.mem_free(); print(current, last-current)
			dotmatrix.clear_dots()
			if ((config.SleepEnd < now < config.SleepStart) if config.SleepEnd < config.SleepStart else not (config.SleepStart < now < config.SleepEnd)):
				dotmatrix.add_dots(language.preHourNumbers[minute5])
				dotmatrix.add_dots(language.dayNumbers[(hour + language.offsets[minute5]) % 24])
				dotmatrix.add_dots(language.postHourNumbers[ minute5])

		# print(f'\r{brightnesssensor.rawBrightness:0.06f} {brightnesssensor.averageBrightness:0.03f} {brightnesssensor.ambientBrightness:0.03f}', end='')
		brightness = brightnesssensor.ambientBrightness if brightnesssensor.ambientBrightness > config.Darkness else 0
		fadeTime = max(0, fadeTime - LOOPTIME)
		dotmatrix.print_dots(brightness, ((config.FadeTime - fadeTime) / config.FadeTime) if config.FadeTime > 0 else 0)
		await uasyncio.sleep_ms(LOOPTIME)

if __name__ == '__main__': uasyncio.run(main())
