import machine, ubinascii, math, config

CLOCK_FREQUENCY = 800000

# WS2812 LedStrip

# rp2 PICO_W
# id=0, sck=Pin(18), mosi=Pin(19), miso=Pin(16)
# id=1, sck=Pin(10), mosi=Pin(11), miso=Pin(8)

# esp32 GENERIC_S3
# id=1, sck=Pin(12), mosi=Pin(11), miso=Pin(13)
# id=2, sck=Pin(18), mosi=Pin(23), miso=Pin(19) ???

# esp8266 GENERIC
# id=1, sck=Pin(14), mosi=Pin(13), miso=Pin(12)

spi = machine.SPI(1, baudrate=CLOCK_FREQUENCY*4) if config.Module == 'esp8266' else machine.SPI(1, baudrate=CLOCK_FREQUENCY*4, sck=machine.Pin(5), mosi=machine.Pin(4), miso=None)
length = max(config.LedNumbers)
strip = bytearray((12 if config.LED == 'SK8612' else 4 if config.LED == 'APA102' else 3) * (length + (0 if config.LED == 'WS2801' else 2)))
if config.LED == 'APA102': strip[-4:] = bytes((0xff, 0xff, 0xff, 0xff))

def rgb2hex(color): return '#' + ubinascii.hexlify(bytes(color)).decode()
def hex2rgb(color): return ubinascii.unhexlify(color[1:])

def clear_strip():
	for i in range(length): colorFrame(i + (0 if config.LED == 'WS2801' else 1), 0)
	write()

def colorFrame(i, brightness):
	if config.LED == 'WS2801': i -= 1
	if not 0 < i <= length: return

	r, g, b = hex2rgb(config.TextColor)
	if config.LED == 'APA102':
		v = math.ceil(brightness * 31)
		a = brightness * (31 / max(v, 1))
		r, g, b = math.ceil(r * a), math.ceil(g * a), math.ceil(b * a)
		strip[i*4:i*4+4] = bytes((0xE0 | v, b, g, r))
		# if brightness > 0: print(f'\r{v:03d}, {b:03d}, {g:03d}, {r:03d}', end='')
	else:
		r, g, b = math.ceil(r * brightness), math.ceil(g * brightness), math.ceil(b * brightness)
		if config.LED == 'SK8612':
			SK6812_LO, SK6812_HI = 0b1000, 0b1100
			strip[i*12+ 0] = (SK6812_HI if g & 0x80 else SK6812_LO) << 4 | (SK6812_HI if g & 0x40 else SK6812_LO)
			strip[i*12+ 1] = (SK6812_HI if g & 0x20 else SK6812_LO) << 4 | (SK6812_HI if g & 0x10 else SK6812_LO)
			strip[i*12+ 2] = (SK6812_HI if g & 0x08 else SK6812_LO) << 4 | (SK6812_HI if g & 0x04 else SK6812_LO)
			strip[i*12+ 3] = (SK6812_HI if g & 0x02 else SK6812_LO) << 4 | (SK6812_HI if g & 0x01 else SK6812_LO)
	
			strip[i*12+ 4] = (SK6812_HI if r & 0x80 else SK6812_LO) << 4 | (SK6812_HI if r & 0x40 else SK6812_LO)
			strip[i*12+ 5] = (SK6812_HI if r & 0x20 else SK6812_LO) << 4 | (SK6812_HI if r & 0x10 else SK6812_LO)
			strip[i*12+ 6] = (SK6812_HI if r & 0x08 else SK6812_LO) << 4 | (SK6812_HI if r & 0x04 else SK6812_LO)
			strip[i*12+ 7] = (SK6812_HI if r & 0x02 else SK6812_LO) << 4 | (SK6812_HI if r & 0x01 else SK6812_LO)
	
			strip[i*12+ 8] = (SK6812_HI if b & 0x80 else SK6812_LO) << 4 | (SK6812_HI if b & 0x40 else SK6812_LO)
			strip[i*12+ 9] = (SK6812_HI if b & 0x20 else SK6812_LO) << 4 | (SK6812_HI if b & 0x10 else SK6812_LO)
			strip[i*12+10] = (SK6812_HI if b & 0x08 else SK6812_LO) << 4 | (SK6812_HI if b & 0x04 else SK6812_LO)
			strip[i*12+11] = (SK6812_HI if b & 0x02 else SK6812_LO) << 4 | (SK6812_HI if b & 0x01 else SK6812_LO)
		elif config.LED == 'WS2801':
			strip[i*3:i*3+3] = bytes((r, g, b))

def write(): spi.write(strip)
